package com.cericlabs.mc.mob_timer;

import java.util.*;
import java.util.logging.Logger;

import org.bukkit.Server;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.LivingEntity;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.entity.CreatureSpawnEvent;
import org.bukkit.event.entity.CreatureSpawnEvent.SpawnReason;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.event.entity.EntityDeathEvent;
import org.bukkit.plugin.PluginManager;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitScheduler;



/**
 * A simple timer that kills mobs that live for longer than a configurable duration.
 * 
 * NOTE: The bukkit API is fucking awful. FUCKING. AWFUL. Some of the code here is not up to the
 * quality I usually hold myself to, but when I'm limited to using a pisspoor interface and event
 * handling scheme such as the one defined here... well... garbage in, garbage out. :/
 * 
 * @author Chris "Ceiu" Rog
 */
public class MobTimer extends JavaPlugin implements Listener {
	
////////////////////////////////////////////////////////////////////////////////////////////////////
	
	private static class EntityLifespanTimer implements Runnable {
		public int id;	// Sigh...
		
		private MobTimer mob_timer;
		private Entity entity;
		
		public long lifespan;
		public boolean expired;
		
		public EntityLifespanTimer(MobTimer mob_timer, Entity entity, long lifespan) {
			if(mob_timer == null)
				throw new IllegalArgumentException("mob_timer");
			
			if(entity == null)
				throw new IllegalArgumentException("entity");
		
			this.id = -1;
			
			this.mob_timer = mob_timer;
			this.entity = entity;
			
			this.lifespan = lifespan;
			this.expired = false;
		}
		
		@Override
		public void run() {
			this.expired = true;
			this.mob_timer.onEntityTimeout(this.entity, this);
		}
	}
	
	private static class EntitySourceConfig {
		public final long default_lifespan;
		public final boolean use_default;
		
		public final Map<EntityType, Long> entity_lifespan;
		
		public EntitySourceConfig() {
			this.default_lifespan = 0;
			this.use_default = false;
			
			this.entity_lifespan = new HashMap<EntityType, Long>(25);
		}
		
		public EntitySourceConfig(long default_lifespan) {
			this.default_lifespan = default_lifespan;
			this.use_default = true;
			
			this.entity_lifespan = new HashMap<EntityType, Long>(25);
		}
	}

////////////////////////////////////////////////////////////////////////////////////////////////////

	// State management...
	private volatile boolean enabled;

	// Objects...
	private Logger logger;
	private BukkitScheduler scheduler;
	
	private final Map<Entity, EntityLifespanTimer> entity_tasks;	// Used for entity->task lookups.
	private final Map<CreatureSpawnEvent.SpawnReason, EntitySourceConfig> entity_sources;	// Lifespan for select entities.
	
	private boolean reset_on_interaction;
	private boolean allow_drops;
	
	
////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public MobTimer() {
		this.entity_tasks = new HashMap<Entity, EntityLifespanTimer>(1000, 1.5f);
		this.entity_sources = new HashMap<CreatureSpawnEvent.SpawnReason, EntitySourceConfig>(25);
		
		this.enabled = false;
		this.reset_on_interaction = false;
		this.allow_drops = false;
	}
	
////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override
	public synchronized void onEnable() {
		if(this.enabled)
			throw new IllegalStateException("MobTimer is already enabled.");
		
		// Get various objects...
		Server server;
		PluginManager pm;
		
		if((server = super.getServer()) == null)
			throw new IllegalStateException("Unable to get server reference.");
				
		if((this.logger = super.getLogger()) == null)
			throw new IllegalStateException("Unable to get logger reference.");

		if((this.scheduler = server.getScheduler()) == null)
			throw new IllegalStateException("Unable to get scheduler reference.");
		
		if((pm = server.getPluginManager()) == null)
			throw new IllegalStateException("Unable to get plugin manager reference.");
		
		
		// Load lifespan configuration...
		FileConfiguration config = super.getConfig();
		EntitySourceConfig source_config;
		String path;
		
		if(config != null) {
			// Add wildcard-source lifespans...
			path = "ANY.ANY.lifespan";
			source_config = config.contains(path) ? new EntitySourceConfig(config.getLong(path, 0)) : new EntitySourceConfig();
			this.entity_sources.put(null, source_config);
			
			for(EntityType type : EntityType.values()) {
				path = ("ANY." + type.toString().toLowerCase() + ".lifespan");
				if(config.contains(path)) source_config.entity_lifespan.put(type, config.getLong(path, 0));
			}
	
			// Add source-specific lifespans...
			for(SpawnReason reason : SpawnReason.values()) {
				// Grab wildcard value...
				path = (reason.toString().toLowerCase() + ".ANY.lifespan");
				source_config = config.contains(path) ? new EntitySourceConfig(config.getLong(path, 0)) : new EntitySourceConfig();
				this.entity_sources.put(reason, source_config);

				// Grab specific values...
				for(EntityType type : EntityType.values()) {
					path = (reason.toString() + "." + type.toString() + ".lifespan").toLowerCase();
					if(config.contains(path)) source_config.entity_lifespan.put(type, config.getLong(path, 0));
				}
			}
			
			// Load other configuration...
			this.reset_on_interaction = config.getBoolean("config.reset_on_interaction", true);
			if(this.reset_on_interaction) logger.info("Resetting lifespan on interaction.");
			
			this.allow_drops = config.getBoolean("config.allow_drops", false);
			if(this.allow_drops) logger.info("Allowing item drops.");
		}
		
		// Create timer and prepare maps...
		this.entity_tasks.clear();
		this.scheduler.cancelTasks(this);
		
		// Register event handlers...
		pm.registerEvents(this, this);
		
		// Done! Enable stuff...
		this.enabled = true;
	}
	
	@Override
	public synchronized void onDisable() {
		if(!this.enabled)
			throw new IllegalStateException("MobTimer is not enabled.");
		
		this.enabled = false;
		
		// Clear timer and maps...
		this.entity_tasks.clear();
		this.scheduler.cancelTasks(this);
		
		// Remove event handlers...
		// Nevermind. I guess you can't do this for some reason. I guess the references will just
		// sit in a list until the server restarts. :/
	}
	

//////////////////////////////////////////////////////////////////////////////////////////////////
	
	private long getLifespan(EntityType type, SpawnReason source) {
		if(type == null)
			throw new IllegalArgumentException("type");
		
		
		EntitySourceConfig source_config = this.entity_sources.get(source);
		Long value;
		
		if(source_config != null) {
			if((value = source_config.entity_lifespan.get(type)) != null)
				return value.longValue();
			
			if(source_config.use_default)
				return source_config.default_lifespan;
		}

		
		EntitySourceConfig wildcard_config = this.entity_sources.get(null);
		
		if(wildcard_config != null) {
			if((value = wildcard_config.entity_lifespan.get(type)) != null)
				return value.longValue();
			
			if(wildcard_config.use_default)
				return wildcard_config.default_lifespan;
		}
		
		
		return 0;
	}
	
//////////////////////////////////////////////////////////////////////////////////////////////////
	
	@EventHandler
	public void onCreatureSpawn(CreatureSpawnEvent event) {
		if(!this.enabled)
			return; // Not enabled. Do nothing.
		
		if(event.isCancelled())
			return; // Canceled event. Creature isn't spawning.
		
		// Get the lifespan of this entity...
		long lifespan = getLifespan(event.getEntityType(), event.getSpawnReason());
		
		if(lifespan > 0) {
			// Get our references...
			Entity entity = event.getEntity();
			EntityLifespanTimer task = new EntityLifespanTimer(this, entity, lifespan);
			
			if((task.id = this.scheduler.scheduleAsyncDelayedTask(this, task, task.lifespan)) != -1) {
				// Store references to our task...
				this.entity_tasks.put(entity, task);
			
				//this.logger.info("Scheduled timer for entity: " + entity + "> " + task.lifespan + " ticks.");
			} else {
				this.logger.warning("Unable to schedule timer for entity: " + entity);
			}
		}
	}
	
	@EventHandler
	public void onEntityDamage(EntityDamageEvent event) {
		if(!this.enabled)
			return; // Not enabled. Do nothing.

		// IMPL Note: A cancelled damage event won't deal damage, but it still indicates something
		// is interacting with them.
		
		if(!this.reset_on_interaction)
			return; // We're not resetting on damage.
		
		Entity entity = event.getEntity();
		EntityLifespanTimer task = this.entity_tasks.get(entity);
		
		if(task != null) {
			// Entity got damaged. Reset timer.
			this.scheduler.cancelTask(task.id);
			
			if((task.id = this.scheduler.scheduleSyncDelayedTask(this, task, task.lifespan)) != -1) {
				//this.logger.info("Reset timer for entity: " + entity);
			} else {
				// Couldn't schedule the task. Remove it.
				this.entity_tasks.remove(entity);
				this.logger.warning("Unable to reschedule timer for entity: " + entity);
			}
		}
	}
	
	@EventHandler
	public void onEntityDeath(EntityDeathEvent event) {
		if(!this.enabled)
			return; // Not enabled. Do nothing.
		
		Entity entity = event.getEntity();
		EntityLifespanTimer task = this.entity_tasks.get(entity);
		
		if(task != null) {
			if(task.expired && !this.allow_drops)
				event.getDrops().clear();
			
			// Entity got killed. Cancel timer & remove reference...
			this.scheduler.cancelTask(task.id);
			this.entity_tasks.remove(entity);
		}
	}
	
	public void onEntityTimeout(Entity entity, EntityLifespanTimer task) {
		if(!this.enabled)
			return; // Not enabled. Do nothing.
		
		if(entity.isDead())
			return; // Already dead.
		
		if(entity instanceof LivingEntity) {
			LivingEntity le = (LivingEntity)entity;
			
			int max_health = le.getMaxHealth();
			le.damage(max_health);
		} else {
			entity.remove();
		}
	}
}
